from __future__ import annotations

import weakref
from typing import TYPE_CHECKING
from weakref import WeakValueDictionary

from slugify import slugify

if TYPE_CHECKING:
    from inspect import FrameInfo
    from typing import Any, Dict, List, Optional, Set
    from types import FrameType

    Stack = List[FrameInfo]


scopes_by_name: WeakValueDictionary[str, Scope] = WeakValueDictionary()
name_sets_by_frame_id: Dict[int, Set[str]] = {}


def construct_name(frame: FrameType, namespace: str = '') -> str:
    name: str = '_dysco_'
    if namespace:
        # Slugify limits characters to those matching the `[^-a-zA-Z0-9]+` regex.
        name += slugify(namespace).replace('-', '_') + '_'
    name += hex(id(frame))
    return name


def destructor(frame_id: int, name: str):
    name_set = name_sets_by_frame_id[frame_id]
    name_set.remove(name)


def find_existing_scope(frame: FrameType, namespace: str = '') -> Optional[Scope]:
    name_set = name_sets_by_frame_id.get(id(frame), set())
    for name in name_set:
        candidate_scope = scopes_by_name.get(name)
        if candidate_scope:
            frame_scope = frame.f_locals.get(candidate_scope.name)
            if candidate_scope is frame_scope and namespace == candidate_scope.namespace:
                return candidate_scope
    return None


def find_parent_scope(scope: Scope, stack: Stack):
    for i, frame_info in enumerate(stack):
        parent_scope = find_existing_scope(frame_info.frame, scope.namespace)
        if parent_scope and parent_scope is not scope:
            return parent_scope, stack[i + 1 :]
    return None, []


class Scope:
    def __init__(self, frame: FrameType, namespace: str = ''):
        # Block calling `__init__()` more than once on a given instance.
        if getattr(self, 'initialized', False):
            return
        self.initialized = True

        self.name = construct_name(frame, namespace)
        self.namespace = namespace
        self.variables: Dict[str, Any] = {}

        scopes_by_name[self.name] = self
        frame.f_locals[self.name] = self
        name_set = name_sets_by_frame_id.get(id(frame), set())
        name_set.add(self.name)
        name_sets_by_frame_id[id(frame)] = name_set
        weakref.finalize(self, destructor, id(frame), self.name)

    def __new__(cls, frame: FrameType, namespace: str = ''):
        # Attempt to find an existing scope for this frame.
        existing_scope = find_existing_scope(frame, namespace)
        if existing_scope:
            return existing_scope

        # Otherwise initialize a new scope.
        return super().__new__(cls)



from dysco.dysco import Dysco

__version__ = '0.0.5'
__author__ = 'Evan Sangaline <evan@intoli.com>'
__description__ = 'Dysco provides configurable dynamic scoping behavior in Python.'
__all__ = ['Dysco', 'g']

g = Dysco()



"""Houses the implementation of the main ``Dysco`` class and project API."""
from __future__ import annotations

import inspect
from typing import TYPE_CHECKING

from dysco.scope import Scope, find_parent_scope

if TYPE_CHECKING:
    from typing import Any


class Dysco:
    def __init__(self, namespace: str = '', read_only: bool = False):
        self.__namespace = namespace
        self.__read_only = read_only

    def __getattr__(self, attribute: str) -> Any:
        if attribute.startswith('_Dysco_'):
            return super().__getattribute__(attribute)
        stack = inspect.stack()
        current_frame = stack.pop(0).frame
        try:
            scope = Scope(stack[0].frame, namespace=self.__namespace)
            while scope:
                if attribute in scope.variables:
                    return scope.variables[attribute]
                scope, stack = find_parent_scope(scope, stack)
            raise AttributeError(f'The attribute {attribute} was not found in any scope.')
        finally:
            # Delete the current frame to avoid reference cycles.
            del current_frame

    def __setattr__(self, attribute: str, value: Any) -> None:
        if attribute.startswith('_Dysco_'):
            super().__setattr__(attribute, value)
        read_only = getattr(self, '_Dysco_read_only', False)
        stack = inspect.stack()
        current_frame = stack.pop(0).frame
        try:
            initial_scope = Scope(stack[0].frame, namespace=self.__namespace)
            scope = initial_scope
            while not read_only and scope:
                if attribute in scope.variables:
                    scope.variables[attribute] = value
                    return
                scope, stack = find_parent_scope(scope, stack)
            initial_scope.variables[attribute] = value
        finally:
            # Delete the current frame to avoid reference cycles.
            del current_frame



"""Utilities related to inspection."""
import inspect
from types import FrameType


def get_calling_frame(depth: int = 0) -> FrameType:
    stack = inspect.stack()
    if len(stack) < depth + 1:
        raise ValueError(f'No frame found at {depth} levels up the stack.')
    return stack[depth + 1].frame



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['dysco']

package_data = \
{'': ['*']}

install_requires = \
['python-slugify>=4.0,<5.0']

setup_kwargs = {
    'name': 'dysco',
    'version': '0.0.5',
    'description': 'Dysco provides configurable dynamic scoping behavior in Python.',
    'long_description': '# Dysco\n\n\n## Development\n\nTo install the dependencies locally, you need [poetry](https://poetry.eustace.io/docs/#installation) to be installed.\nYou can then run\n\n```bash\n# This is optional, but highly recommended.\n# It tells poetry to place the virtual environment in `.venv`.\npoetry config settings.virtualenvs.in-project true\n\n# Install all of the dependencies.\npoetry install\n```\n\nto install the project dependencies.\n\nThe library is tested against Python versions 3.7 and 3.8.\nThese are most easily installed using [pyenv](https://github.com/pyenv/pyenv#installation) with the following command.\n\n```bash\n# Install the supported Python versions.\npyenv install --skip-existing 3.7.5\npyenv install --skip-existing 3.8.0\n```\n\nTesting, linting, and document generation can then be run via [tox](https://tox.readthedocs.io/en/latest/).\nThe bare `tox` command will run everything in all environments, or you can break it down by Python version and task.\nFor example, you could run the individual Python 3.8 tasks manually by running the following.\n\n```bash\n# Install the project dependencies in `.tox/py38/`.\ntox -e py38-init\n\n# Run black, flake8, isort, and mypy.\ntox -e py38-lint\n\n# Run the tests and generate a coverage report.\ntox -e py38-test --coverage\n\n## Build the project documentation.\ntox -e py38-docs\n```\n\n## Deployment\n\nYou first need to configure your credentials with poetry.\n\n```bash\npoetry config http-basic.pypi intoli <pypi-password>\n```\n\nYou can then use invoke to bump the version number, commit the changes, tag the version, and deploy to pypi.\n\n```bash\n# Bumps the patch version and deploys the package.\n# Valid options are major, minor, and patch.\ninvoke bump patch\n```\n',
    'author': 'Evan Sangaline',
    'author_email': 'evan@intoli.com',
    'url': 'https://github.com/intoli/dysco/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
